from pathlib import Path

from pluggy import HookimplMarker
from pluggy import HookspecMarker

spec = hookspec = HookspecMarker("abx")
impl = hookimpl = HookimplMarker("abx")


@hookspec
@hookimpl
def get_system_user() -> str:
    # Beware $HOME may not match current EUID, UID, PUID, SUID, there are edge cases
    # - sudo (EUD != UID != SUID)
    # - running with an autodetected UID based on data dir ownership
    #   but mapping of UID:username is broken because it was created
    #   by a different host system, e.g. 911's $HOME outside of docker
    #   might be /usr/lib/lxd instead of /home/archivebox
    # - running as a user that doens't have a home directory
    # - home directory is set to a path that doesn't exist, or is inside a dir we cant read
    return Path('~').expanduser().name




"""
Hookspec for side-effects that ArchiveBox plugins can trigger.

(e.g. network requests, binary execution, remote API calls, external library calls, etc.)
"""

__package__ = 'abx.archivebox'

import abx


@abx.hookspec
def check_remote_seed_connection(urls, extractor, credentials, created_by):
    pass


@abx.hookspec
def exec_extractor(url, extractor, credentials, config):
    pass




__package__ = 'abx.archivebox'

from typing import Iterable, List
import abc



class BaseSearchBackend(abc.ABC):
    name: str

    @staticmethod
    @abc.abstractmethod
    def index(snapshot_id: str, texts: List[str]):
        return

    @staticmethod
    @abc.abstractmethod
    def flush(snapshot_ids: Iterable[str]):
        return

    @staticmethod
    @abc.abstractmethod
    def search(text: str) -> List[str]:
        raise NotImplementedError("search method must be implemented by subclass")




__package__ = 'abx.archivebox'

import os
import sys
import re
from pathlib import Path
from typing import Type, Tuple, Callable, ClassVar, Dict, Any

import toml
from rich import print

from benedict import benedict
from pydantic import model_validator, TypeAdapter, AliasChoices
from pydantic_settings import BaseSettings, SettingsConfigDict, PydanticBaseSettingsSource
from pydantic_settings.sources import TomlConfigSettingsSource

from pydantic_pkgr import func_takes_args_or_kwargs


from . import toml_util


PACKAGE_DIR = Path(__file__).resolve().parent.parent
DATA_DIR = Path(os.getcwd()).resolve()

ARCHIVEBOX_CONFIG_FILE = DATA_DIR / "ArchiveBox.conf"
ARCHIVEBOX_CONFIG_FILE_BAK = ARCHIVEBOX_CONFIG_FILE.parent / ".ArchiveBox.conf.bak"

AUTOFIXES_HEADER = "[AUTOFIXES]"
AUTOFIXES_SUBHEADER = "# The following config was added automatically to fix problems detected at startup:"

_ALREADY_WARNED_ABOUT_UPDATED_CONFIG = set()


class FlatTomlConfigSettingsSource(TomlConfigSettingsSource):
    """
    A source class that loads variables from a TOML file
    """

    def __init__(
        self,
        settings_cls: type[BaseSettings],
        toml_file: Path | None=None,
    ):
        self.toml_file_path = toml_file or settings_cls.model_config.get("toml_file")
        
        self.nested_toml_data = self._read_files(self.toml_file_path)
        self.toml_data = {}
        for top_level_key, top_level_value in self.nested_toml_data.items():
            if isinstance(top_level_value, dict):
                # value is nested, flatten it
                for key, value in top_level_value.items():
                    self.toml_data[key] = value
            else:
                # value is already flat, just set it as-is
                self.toml_data[top_level_key] = top_level_value
                
        # filter toml_data to only include keys that are defined on this settings_cls
        self.toml_data = {
            key: value
            for key, value in self.toml_data.items()
            if key in settings_cls.model_fields
        }
            
        super(TomlConfigSettingsSource, self).__init__(settings_cls, self.toml_data)


class BaseConfigSet(BaseSettings):
    """
    This is the base class for an ArchiveBox ConfigSet.
    It handles loading values from schema defaults, ArchiveBox.conf TOML config, and environment variables.

    class WgetConfig(ArchiveBoxBaseConfig):
        WGET_BINARY: str = Field(default='wget', alias='WGET_BINARY_PATH')

    c = WgetConfig()
    print(c.WGET_BINARY)                    # outputs: wget

    # you can mutate process environment variable and reload config using .__init__()
    os.environ['WGET_BINARY_PATH'] = 'wget2'
    c.__init__()

    print(c.WGET_BINARY)                    # outputs: wget2

    """
    
    # these pydantic config options are all VERY carefully chosen, make sure to test thoroughly before changing!!!
    model_config = SettingsConfigDict(
        validate_default=False,
        case_sensitive=True,
        extra="ignore",
        arbitrary_types_allowed=False,
        populate_by_name=True,
        from_attributes=True,
        loc_by_alias=False,
        validate_assignment=True,
        validate_return=True,
        revalidate_instances="subclass-instances",
    )
    
    load_from_defaults: ClassVar[bool] = True
    load_from_collection: ClassVar[bool] = True
    load_from_environment: ClassVar[bool] = True

    @classmethod
    def settings_customise_sources(
        cls,
        settings_cls: Type[BaseSettings],
        init_settings: PydanticBaseSettingsSource,
        env_settings: PydanticBaseSettingsSource,
        dotenv_settings: PydanticBaseSettingsSource,
        file_secret_settings: PydanticBaseSettingsSource,
    ) -> Tuple[PydanticBaseSettingsSource, ...]:
        """Defines the config precedence order: Schema defaults -> ArchiveBox.conf (TOML) -> Environment variables"""
        
        # import ipdb; ipdb.set_trace()
        
        precedence_order = {}
        
        # if ArchiveBox.conf does not exist yet, return defaults -> env order
        if not ARCHIVEBOX_CONFIG_FILE.is_file():
            precedence_order = {
                'defaults': init_settings,
                'environment': env_settings,
            }
        
        # if ArchiveBox.conf exists and is in TOML format, return default -> TOML -> env order
        try:
            precedence_order = precedence_order or {
                'defaults': init_settings,
                # 'collection': FlatTomlConfigSettingsSource(settings_cls, toml_file=ARCHIVEBOX_CONFIG_FILE),
                'collection': FlatTomlConfigSettingsSource(settings_cls, toml_file=ARCHIVEBOX_CONFIG_FILE),
                'environment': env_settings,
            }
        except Exception as err:
            if err.__class__.__name__ != "TOMLDecodeError":
                raise
            # if ArchiveBox.conf exists and is in INI format, convert it then return default -> TOML -> env order

            # Convert ArchiveBox.conf in INI format to TOML and save original to .ArchiveBox.bak
            original_ini = ARCHIVEBOX_CONFIG_FILE.read_text()
            ARCHIVEBOX_CONFIG_FILE_BAK.write_text(original_ini)
            new_toml = toml_util.convert(original_ini)
            ARCHIVEBOX_CONFIG_FILE.write_text(new_toml)

            precedence_order = {
                'defaults': init_settings,
                # 'collection': FlatTomlConfigSettingsSource(settings_cls, toml_file=ARCHIVEBOX_CONFIG_FILE),
                'collection': FlatTomlConfigSettingsSource(settings_cls, toml_file=ARCHIVEBOX_CONFIG_FILE),
                'environment': env_settings,
            }
            
        if not cls.load_from_environment:
            precedence_order.pop('environment')
        if not cls.load_from_collection:
            precedence_order.pop('collection')
        if not cls.load_from_defaults:
            precedence_order.pop('defaults')

        return tuple(precedence_order.values())

    @model_validator(mode="after")
    def fill_defaults(self):
        """Populate any unset values using function provided as their default"""

        for key in self.model_fields.keys():
            if isinstance(getattr(self, key), Callable):
                if self.load_from_defaults:
                    computed_default = self.get_default_value(key)
                    # set generated default value as final validated value
                    setattr(self, key, computed_default)
        return self
    
    def validate(self):
        """Manual validation method, to be called from plugin/__init__.py:get_CONFIG()"""
        pass
    
    def get_default_value(self, key):
        """Get the default value for a given config key"""
        field = self.model_fields[key]
        value = getattr(self, key)
    
        if isinstance(value, Callable):
            # if value is a function, execute it to get the actual value, passing existing config as a dict arg if expected
            if func_takes_args_or_kwargs(value):
                # assemble dict of existing field values to pass to default factory functions
                config_so_far = benedict(self.model_dump(include=set(self.model_fields.keys()), warnings=False))
                computed_default = field.default(config_so_far)
            else:
                # otherwise it's a pure function with no args, just call it
                computed_default = field.default()

            # coerce/check to make sure default factory return value matches type annotation
            TypeAdapter(field.annotation).validate_python(computed_default)

            return computed_default
        return value
    
    def update_in_place(self, warn=True, persist=False, hint='', **kwargs):
        """
        Update the config with new values. Use this sparingly! We should almost never be updating config at runtime.
        Sets them in the environment so they propagate to spawned subprocesses / across future re-__init__()s and reload from environment

        Example acceptable use case: user config says SEARCH_BACKEND_ENGINE=sonic but sonic_client pip library is not installed so we cannot use it.
        SEARCH_BACKEND_CONFIG.update_in_place(SEARCH_BACKEND_ENGINE='ripgrep') can be used to reset it back to ripgrep so we can continue.
        """
        from archivebox.misc.toml_util import CustomTOMLEncoder
        
        # silence warnings if they've already been shown once
        if all(key in _ALREADY_WARNED_ABOUT_UPDATED_CONFIG for key in kwargs.keys()):
            warn = False
        
        if warn:
            fix_scope = 'in ArchiveBox.conf' if persist else 'just for current run'
            print(f'\n[yellow]:warning:  WARNING: Some config cannot be used as-is, fixing automatically {fix_scope}:[/yellow] {hint}', file=sys.stderr)
        
        # set the new values in the environment
        for key, value in kwargs.items():
            os.environ[key] = str(value)
            original_value = getattr(self, key)
            if warn:
                print(f'    {key}={original_value} -> {value}')
                _ALREADY_WARNED_ABOUT_UPDATED_CONFIG.add(key)
        
        # if persist=True, write config changes to data/ArchiveBox.conf [AUTOFIXES] section
        try:
            if persist and ARCHIVEBOX_CONFIG_FILE.is_file():
                autofixes_to_add = benedict(kwargs).to_toml(encoder=CustomTOMLEncoder())
                
                existing_config = ARCHIVEBOX_CONFIG_FILE.read_text().split(AUTOFIXES_HEADER, 1)[0].strip()
                if AUTOFIXES_HEADER in existing_config:
                    existing_autofixes = existing_config.split(AUTOFIXES_HEADER, 1)[-1].strip().replace(AUTOFIXES_SUBHEADER, '').replace(AUTOFIXES_HEADER, '').strip()
                else:
                    existing_autofixes = ''
                
                new_config = '\n'.join(line for line in [
                    existing_config,
                    '\n' + AUTOFIXES_HEADER,
                    AUTOFIXES_SUBHEADER,
                    existing_autofixes,
                    autofixes_to_add,
                ] if line.strip()).strip() + '\n'
                ARCHIVEBOX_CONFIG_FILE.write_text(new_config)
        except Exception:
            pass
        self.__init__()
        if warn:
            print(file=sys.stderr)
            
        return self
    
    @property
    def aliases(self) -> Dict[str, str]:
        alias_map = {}
        for key, field in self.model_fields.items():
            alias_map[key] = key
            
            if field.validation_alias is None:
                continue

            if isinstance(field.validation_alias, AliasChoices):
                for alias in field.validation_alias.choices:
                    alias_map[alias] = key
            elif isinstance(field.alias, str):
                alias_map[field.alias] = key
            else:
                raise ValueError(f'Unknown alias type for field {key}: {field.alias}')
        
        return benedict(alias_map)
    
    
    @property
    def toml_section_header(self):
        """Convert the class name to a TOML section header e.g. ShellConfig -> SHELL_CONFIG"""
        class_name = self.__class__.__name__
        return re.sub('([A-Z]+)', r'_\1', class_name).upper().strip('_')
    
    
    def from_defaults(self) -> Dict[str, Any]:
        """Get the dictionary of {key: value} config loaded from the default values"""
        class OnlyDefaultsConfig(self.__class__):
            load_from_defaults = True
            load_from_collection = False
            load_from_environment = False
        return benedict(OnlyDefaultsConfig().model_dump(exclude_unset=False, exclude_defaults=False, exclude=set(self.model_computed_fields.keys())))
    
    def from_collection(self) -> Dict[str, Any]:
        """Get the dictionary of {key: value} config loaded from the collection ArchiveBox.conf"""
        class OnlyConfigFileConfig(self.__class__):
            load_from_defaults = False
            load_from_collection = True
            load_from_environment = False
        return benedict(OnlyConfigFileConfig().model_dump(exclude_unset=True, exclude_defaults=False, exclude=set(self.model_computed_fields.keys())))
    
    def from_environment(self) -> Dict[str, Any]:
        """Get the dictionary of {key: value} config loaded from the environment variables"""
        class OnlyEnvironmentConfig(self.__class__):
            load_from_defaults = False
            load_from_collection = False
            load_from_environment = True
        return benedict(OnlyEnvironmentConfig().model_dump(exclude_unset=True, exclude_defaults=False, exclude=set(self.model_computed_fields.keys())))
    
    def from_computed(self) -> Dict[str, Any]:
        """Get the dictionary of {key: value} config loaded from the computed fields"""
        return benedict(self.model_dump(include=set(self.model_computed_fields.keys())))
    

    def to_toml_dict(self, defaults=False) -> Dict[str, Any]:
        """Get the current config as a TOML-ready dict"""
        config_dict = {}
        for key, value in benedict(self).items():
            if defaults or value != self.get_default_value(key):
                config_dict[key] = value
        
        return benedict({self.toml_section_header: config_dict})
    
    def to_toml_str(self, defaults=False) -> str:
        """Get the current config as a TOML string"""
        from archivebox.misc.toml_util import CustomTOMLEncoder
        
        toml_dict = self.to_toml_dict(defaults=defaults)
        if not toml_dict[self.toml_section_header]:
            # if the section is empty, don't write it
            toml_dict.pop(self.toml_section_header)
        
        return toml.dumps(toml_dict, encoder=CustomTOMLEncoder())
    
    def as_legacy_config_schema(self) -> Dict[str, Any]:
        # shim for backwards compatibility with old config schema style
        model_values = self.model_dump()
        return benedict({
            key: {'type': field.annotation, 'default': model_values[key]}
            for key, field in self.model_fields.items()
        })



__package__ = 'abx.archivebox'

import os
import importlib

from typing import Dict
from pathlib import Path


def load_archivebox_plugins(pm, plugins_dict: Dict[str, Path]):
    """Load archivebox plugins, very similar to abx.load_plugins but it looks for a pydantic PLUGIN model + hooks in apps.py"""
    LOADED_PLUGINS = {}
    for plugin_module, plugin_dir in reversed(plugins_dict.items()):
        # print(f'Loading plugin: {plugin_module} from {plugin_dir}')
        
        # 1. register the plugin module directly in case it contains any look hookimpls (e.g. in __init__.py)
        try:
            plugin_module_loaded = importlib.import_module(plugin_module)
            pm.register(plugin_module_loaded)
        except Exception as e:
            print(f'Error registering plugin: {plugin_module} - {e}')
            
        
        # 2. then try to import plugin_module.apps as well
        if os.access(plugin_dir / 'apps.py', os.R_OK):
            plugin_apps = importlib.import_module(plugin_module + '.apps')
            pm.register(plugin_apps)                                           # register the whole .apps  in case it contains loose hookimpls (not in a class)
            
        # print(f'    √ Loaded plugin: {plugin_module} {len(archivebox_plugins_found) * "🧩"}')
    return LOADED_PLUGINS



"""
Hookspec for ArchiveBox system events that plugins can hook into.

Loosely modeled after Django's signals architecture.
https://docs.djangoproject.com/en/5.1/ref/signals/
"""

__package__ = 'abx.archivebox'

import abx



@abx.hookspec
def on_crawl_schedule_tick(crawl_schedule):
    pass




@abx.hookspec
def on_seed_post_save(seed, created=False):
    ...

@abx.hookspec
def on_crawl_post_save(crawl, created=False):
    ...


@abx.hookspec
def on_snapshot_post_save(snapshot, created=False):
    ...
    
# @abx.hookspec
# def on_snapshot_post_delete(snapshot):
#     ...


@abx.hookspec
def on_archiveresult_post_save(archiveresult, created=False):
    ...

# @abx.hookspec
# def on_archiveresult_post_delete(archiveresult):
#     ...



__package__ = 'abx.archivebox'

import importlib
from typing import Dict, Set, Any, TYPE_CHECKING

from benedict import benedict

from django.conf import settings

import abx
from .. import pm


@abx.hookimpl
def get_or_create_snapshot(crawl, url, config):
    pass

@abx.hookimpl
def update_crawl_schedule_next_run_at(crawl_schedule, next_run_at):
    pass

@abx.hookimpl
def create_crawl_copy(crawl_to_copy, schedule):
    pass

@abx.hookimpl
def create_crawl(seed, depth, tags_str, persona, created_by, config, schedule):
    pass




def create_crawl_from_ui_action(urls, extractor, credentials, depth, tags_str, persona, created_by, crawl_config):
    if seed_is_remote(urls, extractor, credentials):
        # user's seed is a remote source that will provide the urls (e.g. RSS feed URL, Pocket API, etc.)
        uri, extractor, credentials = abx.archivebox.effects.check_remote_seed_connection(urls, extractor, credentials, created_by)
    else:
        # user's seed is some raw text they provided to parse for urls, save it to a file then load the file as a Seed
        uri = abx.archivebox.writes.write_raw_urls_to_local_file(urls, extractor, tags_str, created_by)  # file:///data/sources/some_import.txt
    
    seed = abx.archivebox.writes.get_or_create_seed(uri=remote_uri, extractor, credentials, created_by)
    # abx.archivebox.events.on_seed_created(seed)
        
    crawl = abx.archivebox.writes.create_crawl(seed=seed, depth=depth, tags_str=tags_str, persona=persona, created_by=created_by, config=crawl_config, schedule=None)
    abx.archivebox.events.on_crawl_created(crawl)


@abx.hookimpl(specname='on_crawl_schedule_tick')
def create_crawl_from_crawlschedule_if_due(crawl_schedule):
    # make sure it's not too early to run this scheduled import (makes this function indepmpotent / safe to call multiple times / every second)
    if timezone.now() < crawl_schedule.next_run_at:
        # it's not time to run it yet, wait for the next tick
        return
    else:
        # we're going to run it now, bump the next run time so that no one else runs it at the same time as us
        abx.archivebox.writes.update_crawl_schedule_next_run_at(crawl_schedule, next_run_at=crawl_schedule.next_run_at + crawl_schedule.interval)
    
    crawl_to_copy = None
    try:
        crawl_to_copy = crawl_schedule.crawl_set.first()  # alternatively use .last() to copy most recent crawl instead of very first crawl
    except Crawl.DoesNotExist:
        # there is no template crawl to base the next one off of
        # user must add at least one crawl to a schedule that serves as the template for all future repeated crawls
        return
    
    new_crawl = abx.archivebox.writes.create_crawl_copy(crawl_to_copy=crawl_to_copy, schedule=crawl_schedule)
    abx.archivebox.events.on_crawl_created(new_crawl)


@abx.hookimpl(specname='on_crawl_post_save')
def create_root_snapshot_from_seed(crawl):
    # create a snapshot for the seed URI which kicks off the crawl
    # only a single extractor will run on it, which will produce outlinks which get added back to the crawl
    root_snapshot, created = abx.archivebox.writes.get_or_create_snapshot(crawl=crawl, url=crawl.seed.uri, config={
        'extractors': (
            abx.archivebox.reads.get_extractors_that_produce_outlinks()
            if crawl.seed.extractor == 'auto' else
            [crawl.seed.extractor]
        ),
        **crawl.seed.config,
    })
    if created:
        abx.archivebox.events.on_snapshot_created(root_snapshot)
        abx.archivebox.writes.update_crawl_stats(started_at=timezone.now())


@abx.hookimpl(specname='on_snapshot_created')
def create_archiveresults_pending_from_snapshot(snapshot, config):
    config = get_scope_config(
        # defaults=settings.CONFIG_FROM_DEFAULTS,
        # collection=settings.CONFIG_FROM_FILE,
        # environment=settings.CONFIG_FROM_ENVIRONMENT,
        persona=archiveresult.snapshot.crawl.persona,
        seed=archiveresult.snapshot.crawl.seed,
        crawl=archiveresult.snapshot.crawl,
        snapshot=archiveresult.snapshot,
        archiveresult=archiveresult,
        # extra_config=extra_config,
    )
    
    extractors = abx.archivebox.reads.get_extractors_for_snapshot(snapshot, config)
    for extractor in extractors:
        archiveresult, created = abx.archivebox.writes.get_or_create_archiveresult_pending(
            snapshot=snapshot,
            extractor=extractor,
            status='pending'
        )
        if created:
            abx.archivebox.events.on_archiveresult_created(archiveresult)



@abx.hookimpl(specname='on_archiveresult_updated')
def create_snapshots_pending_from_archiveresult_outlinks(archiveresult):
    config = get_scope_config(...)
    
    # check if extractor has finished succesfully, if not, dont bother checking for outlinks
    if not archiveresult.status == 'succeeded':
        return
    
    # check if we have already reached the maximum recursion depth
    hops_to_here = abx.archivebox.reads.get_outlink_parents(crawl_pk=archiveresult.snapshot.crawl_id, url=archiveresult.url, config=config)
    if len(hops_to_here) >= archiveresult.crawl.max_depth +1:
        return
    
    # parse the output to get outlink url_entries
    discovered_urls = abx.archivebox.reads.get_archiveresult_discovered_url_entries(archiveresult, config=config)
    
    for url_entry in discovered_urls:
        abx.archivebox.writes.create_outlink_record(src=archiveresult.snapshot.url, dst=url_entry.url, via=archiveresult)
        abx.archivebox.writes.create_snapshot(crawl=archiveresult.snapshot.crawl, url_entry=url_entry)
        
    # abx.archivebox.events.on_crawl_updated(archiveresult.snapshot.crawl)



__package__ = 'archivebox.crawls'

import time

import abx
import abx.archivebox.events
import abx.hookimpl

from datetime import datetime

from django_stubs_ext.db.models import TypedModelMeta

from django.db import models
from django.db.models import Q
from django.core.validators import MaxValueValidator, MinValueValidator 
from django.conf import settings
from django.utils import timezone
from django.utils.functional import cached_property
from django.urls import reverse_lazy

from pathlib import Path









# reads

def tick_core():
    tick_crawls()
    tick_snapshots()
    tick_archiveresults()
    time.sleep(0.1)

#################################################################################################################

# [-> queued] -> started -> sealed

SNAPSHOT_STATES = ('queued', 'started', 'sealed')
SNAPSHOT_FINAL_STATES = ('sealed',)


def get_snapshots_queue():
    retry_at_reached = Q(retry_at__isnull=True) | Q(retry_at__lte=time.now())
    not_in_final_state = ~Q(status__in=SNAPSHOT_FINAL_STATES)
    queue = Snapshot.objects.filter(retry_at_reached & not_in_final_state)
    return queue

@djhuey.task(schedule=djhuey.Periodic(seconds=1))
def tick_snapshots():
    queue = get_snapshots_queue()
    try:
        snapshot = queue.last()
        print(f'QUEUE LENGTH: {queue.count()}, PROCESSING SNAPSHOT[{snapshot.status}]: {snapshot}')
        tick_snapshot(snapshot, cwd=snapshot.cwd)
    except Snapshot.DoesNotExist:
        pass


def tick_snapshot(snapshot, config, cwd):
    # [-> queued] -> started -> sealed

    # SEALED (final state, do nothing)
    if snapshot.status in SNAPSHOT_FINAL_STATES:
        assert snapshot.retry_at is None
        return None
    else:
        assert snapshot.retry_at is not None

    # QUEUED -> PARTIAL
    elif snapshot.status == 'queued':
        transition_snapshot_to_started(snapshot, config, cwd)
    
    # PARTIAL -> SEALED
    elif snapshot.status == 'started':
        if snapshot_has_pending_archiveresults(snapshot, config, cwd):
            # tasks still in-progress, check back again in another 5s
            snapshot.retry_at = time.now() + timedelta(seconds=5)
            snapshot.save()
        else:
            # everything is finished, seal the snapshot
            transition_snapshot_to_sealed(snapshot, config, cwd)
            
    update_snapshot_index_json(archiveresult, config, cwd)
    update_snapshot_index_html(archiveresult, config, cwd)


def transition_snapshot_to_started(snapshot, config, cwd):
    # queued [-> started] -> sealed
    
    retry_at = time.now() + timedelta(seconds=10)
    retries = snapshot.retries + 1
    
    snapshot_to_update = {'pk': snapshot.pk, 'status': 'queued'}
    fields_to_update = {'status': 'started', 'retry_at': retry_at, 'retries': retries, 'start_ts': time.now(), 'end_ts': None}
    snapshot = abx.archivebox.writes.update_snapshot(filter_kwargs=snapshot_to_update, update_kwargs=fields_to_update)
    
    cleanup_snapshot_dir(snapshot, config, cwd)
    create_snapshot_pending_archiveresults(snapshot, config, cwd)
    update_snapshot_index_json(archiveresult, config, cwd)
    update_snapshot_index_html(archiveresult, config, cwd)
    
    


def transition_snapshot_to_sealed(snapshot, config, cwd):
    # -> queued -> started [-> sealed]
    
    snapshot_to_update = {'pk': snapshot.pk, 'status': 'started'}
    fields_to_update = {'status': 'sealed', 'retry_at': None, 'end_ts': time.now()}
    snapshot = abx.archivebox.writes.update_snapshot(filter_kwargs=snapshot_to_update, update_kwargs=fields_to_update)

    cleanup_snapshot_dir(snapshot, config, cwd)
    update_snapshot_index_json(snapshot, config, cwd)
    update_snapshot_index_html(snapshot, config, cwd)
    seal_snapshot_dir(snapshot, config, cwd)  # generate merkle tree and sign the snapshot
    upload_snapshot_dir(snapshot, config, cwd)  # upload to s3, ipfs, etc
    return snapshot


def tick_crawl(crawl, config, cwd):
    # [-> pending] -> archiving -> sealed
    pass


@abx.hookimpl
def create_queued_archiveresult_on_snapshot(snapshot, config) -> bool | None:
    # [-> queued] -> started -> succeeded
    #                        -> backoff   -> queued
    #                                     -> failed
    if not config.SAVE_WARC:
        return None
    
    existing_results = abx.archivebox.reads.get_archiveresults_from_snapshot(snapshot, extractor='warc')
    has_pending_or_succeeded_results = any(result.status in ('queued', 'started', 'succeeded', 'backoff') for result in existing_results)
    if not has_pending_or_succeeded_results:
        return abx.archivebox.writes.create_archiveresult(snapshot=snapshot, extractor='warc', status='queued', retry_at=time.now())
    return None


#################################################################################################################

# [-> queued] -> started -> succeeded
#                        -> backoff   -> queued
#                                     -> failed

ARCHIVERESULT_STATES = ('queued', 'started', 'succeeded', 'backoff', 'failed')
ARCHIVERESULT_FINAL_STATES = ('succeeded', 'failed')


def get_archiveresults_queue():
    retry_at_reached = Q(retry_at__isnull=True) | Q(retry_at__lte=time.now())
    not_in_final_state = ~Q(status__in=ARCHIVERESULT_FINAL_STATES)
    queue = ArchiveResult.objects.filter(retry_at_reached & not_in_final_state)
    return queue

@djhuey.task(schedule=djhuey.Periodic(seconds=1))
def tick_archiveresults():
    queue = get_archiveresults_queue()
    try:
        archiveresult = queue.last()
        print(f'QUEUE LENGTH: {queue.count()}, PROCESSING {archiveresult.status} ARCHIVERESULT: {archiveresult}')
        tick_archiveresult(archiveresult, cwd=archiveresult.cwd)
    except ArchiveResult.DoesNotExist:
        pass

def tick_archiveresult(archiveresult, cwd):
    # [-> queued] -> started -> succeeded
    #                        -> backoff   -> queued
    #                                     -> failed
    
    start_state = archiveresult.status

    # SUCCEEDED or FAILED (final state, do nothing)
    if archiveresult.status in ARCHIVERESULT_FINAL_STATES:
        return None
    
    # QUEUED -> STARTED
    elif archiveresult.status == 'queued':
        transition_archiveresult_to_started(archiveresult, config, cwd)
    
    # STARTED -> SUCCEEDED or BACKOFF
    elif archiveresult.status == 'started':
        if check_if_extractor_succeeded(archiveresult, config, cwd):
            transition_archiveresult_to_succeeded(archiveresult, config, cwd)
        else:
            transition_archiveresult_to_backoff(archiveresult, config, cwd)

    # BACKOFF -> QUEUED or FAILED
    elif archiveresult.status == 'backoff':
        if too_many_retries(archiveresult, config):
            transition_archiveresult_to_failed(archiveresult, config, cwd)
        else:
            transition_archiveresult_to_queued(archiveresult, config, cwd)
            
    end_state = archiveresult.status
    
    # trigger a tick on the Snapshot as well
    archiveresult.snapshot.retry_at = time.now()
    archiveresult.snapshot.save()

    # trigger side effects on state transitions, e.g.:
    #     queued -> started: create the extractor output dir, load extractor binary, spawn the extractor subprocess
    #     started -> succeeded: cleanup the extractor output dir and move into snapshot.link_dir, write index.html, index.json, write logs
    #     started -> backoff: cleanup the extractor output dir, wrtie index.html, index.json collect stdout/stderr logs
    #     backoff -> queued: spawn the extractor subprocess later
    #     *       -> *:      write index.html, index.json, bump ArchiveResult.updated and Snapshot.updated timestamps


def transition_archiveresult_to_started(archiveresult, config, cwd):
    # queued [-> started] -> succeeded
    #                     -> backoff   -> queued
    #                                  -> failed
    
    from .extractors import WARC_EXTRACTOR
    
    # ok, a warc ArchiveResult is queued, let's try to claim it
    retry_at = time.now() + timedelta(seconds=config.TIMEOUT + 5)   # add 5sec buffer so we dont retry things if the previous task is doing post-task cleanup/saving thats taking a little longer than usual
    retries = archiveresult.retries + 1
    archiveresult_to_update = {'pk': archiveresult.pk, 'status': 'queued'}
    fields_to_update = {'status': 'started', 'retry_at': retry_at, 'retries': retries, 'start_ts': time.now(), 'output': None, 'error': None}
    archiveresult = abx.archivebox.writes.update_archiveresult(filter=archiveresult_to_update, update=fields_to_update)
    
    
    with TimedProgress():
        try:
            from .extractors import WARC_EXTRACTOR
            WARC_EXTRACTOR.cleanup_output_dir(archiveresult)
            WARC_EXTRACTOR.load_extractor_binary(archiveresult)
            WARC_EXTRACTOR.extract(archiveresult, config, cwd=archiveresult.cwd)
        except Exception as e:
            WARC_EXTRACTOR.save_error(archiveresult, e)
        finally:
            archiveresult_to_update = {'pk': archiveresult.pk, **fields_to_update}
            fields_to_update = {'retry_at': time.now()}
            archiveresult = abx.archivebox.writes.update_archiveresult(filter_kwargs=archiveresult_to_update, update_kwargs=fields_to_update)
    
    return archiveresult


def transition_archiveresult_to_succeeded(archiveresult, config, cwd):
    output = abx.archivebox.reads.get_archiveresult_output(archiveresult)
    end_ts = time.now()
    
    archiveresult_to_update = {'pk': archiveresult.pk, 'status': 'started'}
    fields_to_update = {'status': 'succeeded', 'retry_at': None, 'end_ts': end_ts, 'output': output}
    archiveresult = abx.archivebox.writes.update_archiveresult(filter_kwargs=archiveresult_to_update, update_kwargs=fields_to_update)
    return archiveresult


def transition_archiveresult_to_backoff(archiveresult, config, cwd):
    # queued -> started [-> backoff]   -> queued
    #                                  -> failed
    #                    -> succeeded
    
    error = abx.archivebox.reads.get_archiveresult_error(archiveresult, cwd)
    end_ts = time.now()
    output = None
    retry_at = time.now() + timedelta(seconds=config.TIMEOUT * archiveresult.retries)
    
    archiveresult_to_update = {'pk': archiveresult.pk, 'status': 'started'}
    fields_to_update = {'status': 'backoff', 'retry_at': retry_at, 'end_ts': end_ts, 'output': output, 'error': error}
    archiveresult = abx.archivebox.writes.update_archiveresult(filter_kwargs=archiveresult_to_update, update_kwargs=fields_to_update)
    return archiveresult


def transition_archiveresult_to_queued(archiveresult, config, cwd):
    # queued -> started -> backoff   [-> queued]
    #                                 -> failed
    #                   -> succeeded
    
    archiveresult_to_update = {'pk': archiveresult.pk, 'status': 'backoff'}
    fields_to_update = {'status': 'queued', 'retry_at': time.now(), 'start_ts': None, 'end_ts': None, 'output': None, 'error': None}
    archiveresult = abx.archivebox.writes.update_archiveresult(filter_kwargs=archiveresult_to_update, update_kwargs=fields_to_update)
    return archiveresult


def transition_archiveresult_to_failed(archiveresult, config, cwd):
    # queued -> started -> backoff    -> queued
    #                                [-> failed]
    #                   -> succeeded
    
    archiveresult_to_update = {'pk': archiveresult.pk, 'status': 'backoff'}
    fields_to_update = {'status': 'failed', 'retry_at': None}
    archiveresult = abx.archivebox.writes.update_archiveresult(filter_kwargs=archiveresult_to_update, update_kwargs=fields_to_update)
    return archiveresult





def should_extract_wget(snapshot, extractor, config) -> bool | None:
    if extractor == 'wget':
        from .extractors import WGET_EXTRACTOR
        return WGET_EXTRACTOR.should_extract(snapshot, config)

def extrac_wget(uri, config, cwd):
    from .extractors import WGET_EXTRACTOR
    return WGET_EXTRACTOR.extract(uri, config, cwd)


@abx.hookimpl
def ready():
    from .config import WGET_CONFIG
    WGET_CONFIG.validate()

















@abx.hookimpl
def on_crawl_schedule_tick(crawl_schedule):
    create_crawl_from_crawl_schedule_if_due(crawl_schedule)

@abx.hookimpl
def on_crawl_created(crawl):
    create_root_snapshot(crawl)

@abx.hookimpl
def on_snapshot_created(snapshot, config):
    create_archiveresults_pending_from_snapshot(snapshot, config)

# events
@abx.hookimpl
def on_archiveresult_created(archiveresult):
    abx.archivebox.exec.exec_archiveresult_extractor(archiveresult)

@abx.hookimpl
def on_archiveresult_updated(archiveresult):
    abx.archivebox.writes.create_snapshots_pending_from_archiveresult_outlinks(archiveresult)




def scheduler_runloop():
    # abx.archivebox.events.on_scheduler_runloop_start(timezone.now(), machine=Machine.objects.get_current_machine())

    while True:
        # abx.archivebox.events.on_scheduler_tick_start(timezone.now(), machine=Machine.objects.get_current_machine())
        
        scheduled_crawls = CrawlSchedule.objects.filter(is_enabled=True)
        scheduled_crawls_due = scheduled_crawls.filter(next_run_at__lte=timezone.now())
        
        for scheduled_crawl in scheduled_crawls_due:
            try:
                abx.archivebox.events.on_crawl_schedule_tick(scheduled_crawl)
            except Exception as e:
                abx.archivebox.events.on_crawl_schedule_failure(timezone.now(), machine=Machine.objects.get_current_machine(), error=e, schedule=scheduled_crawl)
        
        # abx.archivebox.events.on_scheduler_tick_end(timezone.now(), machine=Machine.objects.get_current_machine(), tasks=scheduled_tasks_due)
        time.sleep(1)


def create_crawl_from_ui_action(urls, extractor, credentials, depth, tags_str, persona, created_by, crawl_config):
    if seed_is_remote(urls, extractor, credentials):
        # user's seed is a remote source that will provide the urls (e.g. RSS feed URL, Pocket API, etc.)
        uri, extractor, credentials = abx.archivebox.effects.check_remote_seed_connection(urls, extractor, credentials, created_by)
    else:
        # user's seed is some raw text they provided to parse for urls, save it to a file then load the file as a Seed
        uri = abx.archivebox.writes.write_raw_urls_to_local_file(urls, extractor, tags_str, created_by)  # file:///data/sources/some_import.txt
    
    seed = abx.archivebox.writes.get_or_create_seed(uri=remote_uri, extractor, credentials, created_by)
    # abx.archivebox.events.on_seed_created(seed)
        
    crawl = abx.archivebox.writes.create_crawl(seed=seed, depth=depth, tags_str=tags_str, persona=persona, created_by=created_by, config=crawl_config, schedule=None)
    abx.archivebox.events.on_crawl_created(crawl)


def create_crawl_from_crawl_schedule_if_due(crawl_schedule):
    # make sure it's not too early to run this scheduled import (makes this function indepmpotent / safe to call multiple times / every second)
    if timezone.now() < crawl_schedule.next_run_at:
        # it's not time to run it yet, wait for the next tick
        return
    else:
        # we're going to run it now, bump the next run time so that no one else runs it at the same time as us
        abx.archivebox.writes.update_crawl_schedule_next_run_at(crawl_schedule, next_run_at=crawl_schedule.next_run_at + crawl_schedule.interval)
    
    crawl_to_copy = None
    try:
        crawl_to_copy = crawl_schedule.crawl_set.first()  # alternatively use .last() to copy most recent crawl instead of very first crawl
    except Crawl.DoesNotExist:
        # there is no template crawl to base the next one off of
        # user must add at least one crawl to a schedule that serves as the template for all future repeated crawls
        return
    
    new_crawl = abx.archivebox.writes.create_crawl_copy(crawl_to_copy=crawl_to_copy, schedule=crawl_schedule)
    abx.archivebox.events.on_crawl_created(new_crawl)



def create_root_snapshot(crawl):
    # create a snapshot for the seed URI which kicks off the crawl
    # only a single extractor will run on it, which will produce outlinks which get added back to the crawl
    root_snapshot, created = abx.archivebox.writes.get_or_create_snapshot(crawl=crawl, url=crawl.seed.uri, config={
        'extractors': (
            abx.archivebox.reads.get_extractors_that_produce_outlinks()
            if crawl.seed.extractor == 'auto' else
            [crawl.seed.extractor]
        ),
        **crawl.seed.config,
    })
    if created:
        abx.archivebox.events.on_snapshot_created(root_snapshot)
        abx.archivebox.writes.update_crawl_stats(started_at=timezone.now())


def create_archiveresults_pending_from_snapshot(snapshot, config):
    config = get_scope_config(
        # defaults=settings.CONFIG_FROM_DEFAULTS,
        # configfile=settings.CONFIG_FROM_FILE,
        # environment=settings.CONFIG_FROM_ENVIRONMENT,
        persona=archiveresult.snapshot.crawl.persona,
        seed=archiveresult.snapshot.crawl.seed,
        crawl=archiveresult.snapshot.crawl,
        snapshot=archiveresult.snapshot,
        archiveresult=archiveresult,
        # extra_config=extra_config,
    )
    
    extractors = abx.archivebox.reads.get_extractors_for_snapshot(snapshot, config)
    for extractor in extractors:
        archiveresult, created = abx.archivebox.writes.get_or_create_archiveresult_pending(
            snapshot=snapshot,
            extractor=extractor,
            status='pending'
        )
        if created:
            abx.archivebox.events.on_archiveresult_created(archiveresult)


def exec_archiveresult_extractor(archiveresult):
    config = get_scope_config(...)
    
    # abx.archivebox.writes.update_archiveresult_started(archiveresult, start_ts=timezone.now())
    # abx.archivebox.events.on_archiveresult_updated(archiveresult)
    
    # check if it should be skipped
    if not abx.archivebox.reads.get_archiveresult_should_run(archiveresult, config):
        abx.archivebox.writes.update_archiveresult_skipped(archiveresult, status='skipped')
        abx.archivebox.events.on_archiveresult_skipped(archiveresult, config)
        return
    
    # run the extractor method and save the output back to the archiveresult
    try:
        output = abx.archivebox.writes.exec_archiveresult_extractor(archiveresult, config)
        abx.archivebox.writes.update_archiveresult_succeeded(archiveresult, output=output, error=None, end_ts=timezone.now())
    except Exception as e:
        abx.archivebox.writes.update_archiveresult_failed(archiveresult, error=e, end_ts=timezone.now())
    
    # bump the modified time on the archiveresult and Snapshot
    abx.archivebox.events.on_archiveresult_updated(archiveresult)
    abx.archivebox.events.on_snapshot_updated(archiveresult.snapshot)
    

def create_snapshots_pending_from_archiveresult_outlinks(archiveresult):
    config = get_scope_config(...)
    
    # check if extractor has finished succesfully, if not, dont bother checking for outlinks
    if not archiveresult.status == 'succeeded':
        return
    
    # check if we have already reached the maximum recursion depth
    hops_to_here = abx.archivebox.reads.get_outlink_parents(crawl_pk=archiveresult.snapshot.crawl_id, url=archiveresult.url, config=config)
    if len(hops_to_here) >= archiveresult.crawl.max_depth +1:
        return
    
    # parse the output to get outlink url_entries
    discovered_urls = abx.archivebox.reads.get_archiveresult_discovered_url_entries(archiveresult, config=config)
    
    for url_entry in discovered_urls:
        abx.archivebox.writes.create_outlink_record(src=archiveresult.snapshot.url, dst=url_entry.url, via=archiveresult)
        abx.archivebox.writes.create_snapshot(crawl=archiveresult.snapshot.crawl, url_entry=url_entry)
        
    # abx.archivebox.events.on_crawl_updated(archiveresult.snapshot.crawl)

@abx.hookimpl.reads.get_outlink_parents
def get_outlink_parents(url, crawl_pk=None, config=None):
    scope = Q(dst=url)
    if crawl_pk:
        scope = scope | Q(via__snapshot__crawl_id=crawl_pk)
    
    parent = list(Outlink.objects.filter(scope))
    if not parent:
        # base case: we reached the top of the chain, no more parents left
        return []
    
    # recursive case: there is another parent above us, get its parents
    yield parent[0]
    yield from get_outlink_parents(parent[0].src, crawl_pk=crawl_pk, config=config)





__package__ = 'abx.archivebox'

import importlib
from typing import Dict, Set, Any, TYPE_CHECKING

from benedict import benedict

import abx
from .. import pm

if TYPE_CHECKING:
    from .base_configset import BaseConfigSet
    from .base_binary import BaseBinary, BaseBinProvider
    from .base_extractor import BaseExtractor
    from .base_searchbackend import BaseSearchBackend
    # from .base_replayer import BaseReplayer
    # from .base_queue import BaseQueue
    # from .base_admindataview import BaseAdminDataView

# API exposed to ArchiveBox code

def get_PLUGINS() -> Dict[str, Dict[str, Any]]:
    return benedict({
        plugin_id: plugin
        for plugin_dict in pm.hook.get_PLUGIN()
            for plugin_id, plugin in plugin_dict.items()
    })

def get_PLUGIN(plugin_id: str) -> Dict[str, Any]:
    plugin_info = get_PLUGINS().get(plugin_id, {})
    package = plugin_info.get('package', plugin_info.get('PACKAGE', None))
    if not package:
        return {'id': plugin_id, 'hooks': {}}
    module = importlib.import_module(package)
    hooks = abx.get_plugin_hooks(module.__package__)
    assert plugin_info and (plugin_info.get('id') or plugin_info.get('ID') or hooks)
    
    return benedict({
        'id': plugin_id,
        'label': getattr(module, '__label__', plugin_id),
        'module': module,
        'package': module.__package__,
        'hooks': hooks,
        'version': getattr(module, '__version__', '999.999.999'),
        'author': getattr(module, '__author__', 'Unknown'),
        'homepage': getattr(module, '__homepage__', 'https://github.com/ArchiveBox/ArchiveBox'),
        'dependencies': getattr(module, '__dependencies__', []),
        'source_code': module.__file__,
        **plugin_info,
    })
    

def get_HOOKS() -> Set[str]:
    return {
        hook_name
        for plugin_id in get_PLUGINS().keys()
            for hook_name in get_PLUGIN(plugin_id).hooks
    }

def get_CONFIGS() -> benedict:   # Dict[str, 'BaseConfigSet']
    return benedict({
        config_id: configset
        for plugin_configs in pm.hook.get_CONFIG()
            for config_id, configset in plugin_configs.items()
    })


def get_FLAT_CONFIG() -> Dict[str, Any]:
    return benedict({
        key: value
        for configset in get_CONFIGS().values()
            for key, value in configset.model_dump().items()
    })

def get_BINPROVIDERS() -> Dict[str, 'BaseBinProvider']:
    # TODO: move these to plugins
    from abx.archivebox.base_binary import apt, brew, env
    builtin_binproviders = {
        'env': env,
        'apt': apt,
        'brew': brew,
    }
    
    return benedict({
        binprovider_id: binprovider
        for plugin_binproviders in [builtin_binproviders, *pm.hook.get_BINPROVIDERS()]
            for binprovider_id, binprovider in plugin_binproviders.items()
    })

def get_BINARIES() -> Dict[str, 'BaseBinary']:
    return benedict({
        binary_id: binary
        for plugin_binaries in pm.hook.get_BINARIES()
            for binary_id, binary in plugin_binaries.items()
    })

def get_EXTRACTORS() -> Dict[str, 'BaseExtractor']:
    return benedict({
        extractor_id: extractor
        for plugin_extractors in pm.hook.get_EXTRACTORS()
            for extractor_id, extractor in plugin_extractors.items()
    })

# def get_REPLAYERS() -> Dict[str, 'BaseReplayer']:
#     return benedict({
#         replayer.id: replayer
#         for plugin_replayers in pm.hook.get_REPLAYERS()
#             for replayer in plugin_replayers
#     })

# def get_ADMINDATAVIEWS() -> Dict[str, 'BaseAdminDataView']:
#     return benedict({
#         admin_dataview.id: admin_dataview
#         for plugin_admin_dataviews in pm.hook.get_ADMINDATAVIEWS()
#             for admin_dataview in plugin_admin_dataviews
#     })

# def get_QUEUES() -> Dict[str, 'BaseQueue']:
#     return benedict({
#         queue.id: queue
#         for plugin_queues in pm.hook.get_QUEUES()
#             for queue in plugin_queues
#     })

def get_SEARCHBACKENDS() -> Dict[str, 'BaseSearchBackend']:
    return benedict({
        searchbackend_id: searchbackend
        for plugin_searchbackends in pm.hook.get_SEARCHBACKENDS()
            for searchbackend_id,searchbackend in plugin_searchbackends.items()
    })



def get_scope_config(defaults: benedict | None = None, persona=None, seed=None, crawl=None, snapshot=None, archiveresult=None, extra_config=None):
    """Get all the relevant config for the given scope, in correct precedence order"""
    
    from django.conf import settings
    default_config: benedict = defaults or settings.CONFIG
    
    snapshot = snapshot or (archiveresult and archiveresult.snapshot)
    crawl = crawl or (snapshot and snapshot.crawl)
    seed = seed or (crawl and crawl.seed)
    persona = persona or (crawl and crawl.persona)
    
    persona_config = persona.config if persona else {}
    seed_config = seed.config if seed else {}
    crawl_config = crawl.config if crawl else {}
    snapshot_config = snapshot.config if snapshot else {}
    archiveresult_config = archiveresult.config if archiveresult else {}
    extra_config = extra_config or {}
    
    return {
        **default_config,               # defaults / config file / environment variables
        **persona_config,               # lowest precedence
        **seed_config,
        **crawl_config,
        **snapshot_config,
        **archiveresult_config,
        **extra_config,                 # highest precedence
    }



import inspect

import pluggy


class PluginManager(pluggy.PluginManager):
    """
    Patch to fix pluggy's PluginManager to work with pydantic models.
    See: https://github.com/pytest-dev/pluggy/pull/536
    """
    def parse_hookimpl_opts(self, plugin, name: str) -> pluggy.HookimplOpts | None:
        # IMPORTANT: @property methods can have side effects, and are never hookimpl
        # if attr is a property, skip it in advance
        plugin_class = plugin if inspect.isclass(plugin) else type(plugin)
        if isinstance(getattr(plugin_class, name, None), property):
            return None

        # pydantic model fields are like attrs and also can never be hookimpls
        plugin_is_pydantic_obj = hasattr(plugin, "__pydantic_core_schema__")
        if plugin_is_pydantic_obj and name in getattr(plugin, "model_fields", {}):
            # pydantic models mess with the class and attr __signature__
            # so inspect.isroutine(...) throws exceptions and cant be used
            return None
        
        try:
            return super().parse_hookimpl_opts(plugin, name)
        except AttributeError:
            return super().parse_hookimpl_opts(type(plugin), name)

pm = PluginManager("abx")



__package__ = 'abx.django'

from ..hookspec import hookspec


###########################################################################################

@hookspec
def get_INSTALLED_APPS():
    """Return a list of apps to add to INSTALLED_APPS"""
    # e.g. ['your_plugin_type.plugin_name']
    return []

# @hookspec
# def register_INSTALLED_APPS(INSTALLED_APPS):
#     """Mutate INSTALLED_APPS in place to add your app in a specific position"""
#     # idx_of_contrib = INSTALLED_APPS.index('django.contrib.auth')
#     # INSTALLED_APPS.insert(idx_of_contrib + 1, 'your_plugin_type.plugin_name')
#     pass


@hookspec
def get_TEMPLATE_DIRS():
    return []     # e.g. ['your_plugin_type/plugin_name/templates']

# @hookspec
# def register_TEMPLATE_DIRS(TEMPLATE_DIRS):
#     """Install django settings"""
#     # e.g. TEMPLATE_DIRS.insert(0, 'your_plugin_type/plugin_name/templates')
#     pass


@hookspec
def get_STATICFILES_DIRS():
    return []     # e.g. ['your_plugin_type/plugin_name/static']

# @hookspec
# def register_STATICFILES_DIRS(STATICFILES_DIRS):
#     """Mutate STATICFILES_DIRS in place to add your static dirs in a specific position"""
#     # e.g. STATICFILES_DIRS.insert(0, 'your_plugin_type/plugin_name/static')
#     pass


@hookspec
def get_MIDDLEWARE():
    return []     # e.g. ['your_plugin_type.plugin_name.middleware.YourMiddleware']

# @hookspec
# def register_MIDDLEWARE(MIDDLEWARE):
#     """Mutate MIDDLEWARE in place to add your middleware in a specific position"""
#     # e.g. MIDDLEWARE.insert(0, 'your_plugin_type.plugin_name.middleware.YourMiddleware')
#     pass


@hookspec
def get_AUTHENTICATION_BACKENDS():
    return []     # e.g. ['django_auth_ldap.backend.LDAPBackend']

# @hookspec
# def register_AUTHENTICATION_BACKENDS(AUTHENTICATION_BACKENDS):
#     """Mutate AUTHENTICATION_BACKENDS in place to add your auth backends in a specific position"""
#     # e.g. AUTHENTICATION_BACKENDS.insert(0, 'your_plugin_type.plugin_name.backend.YourBackend')
#     pass

@hookspec
def get_DJANGO_HUEY_QUEUES(QUEUE_DATABASE_NAME):
    return []     # e.g. [{'name': 'your_plugin_type.plugin_name', 'HUEY': {...}}]

# @hookspec
# def register_DJANGO_HUEY(DJANGO_HUEY):
#     """Mutate DJANGO_HUEY in place to add your huey queues in a specific position"""
#     # e.g. DJANGO_HUEY['queues']['some_queue_name']['some_setting'] = 'some_value'
#     pass


@hookspec
def get_ADMIN_DATA_VIEWS_URLS():
    return []

# @hookspec
# def register_ADMIN_DATA_VIEWS(ADMIN_DATA_VIEWS):
#     """Mutate ADMIN_DATA_VIEWS in place to add your admin data views in a specific position"""
#     # e.g. ADMIN_DATA_VIEWS['URLS'].insert(0, 'your_plugin_type/plugin_name/admin_data_views.py')
#     pass


# @hookspec
# def register_settings(settings):
#     """Mutate settings in place to add your settings / modify existing settings"""
#     # settings.SOME_KEY = 'some_value'
#     pass


###########################################################################################

@hookspec
def get_urlpatterns():
    return []     # e.g. [path('your_plugin_type/plugin_name/url.py', your_view)]

# @hookspec
# def register_urlpatterns(urlpatterns):
#     """Mutate urlpatterns in place to add your urlpatterns in a specific position"""
#     # e.g. urlpatterns.insert(0, path('your_plugin_type/plugin_name/url.py', your_view))
#     pass

###########################################################################################

@hookspec
def register_checks():
    """Register django checks with django system checks system"""
    pass

@hookspec
def register_admin(admin_site):
    """Register django admin views/models with the main django admin site instance"""
    pass


###########################################################################################


@hookspec
def ready():
    """Called when Django apps app.ready() are triggered"""
    pass



__package__ = 'abx.archivebox'

from typing import Dict, Any

from .. import hookspec

from .base_binary import BaseBinary, BaseBinProvider
from .base_configset import BaseConfigSet
from .base_extractor import BaseExtractor
from .base_searchbackend import BaseSearchBackend


@hookspec
def get_PLUGIN() -> Dict[str, Dict[str, Any]]:
    return {}

@hookspec
def get_CONFIG() -> Dict[str, BaseConfigSet]:
    return {}



@hookspec
def get_EXTRACTORS() -> Dict[str, BaseExtractor]:
    return {}

@hookspec
def get_SEARCHBACKENDS() -> Dict[str, BaseSearchBackend]:
    return {}

# @hookspec
# def get_REPLAYERS() -> Dict[str, BaseReplayer]:
#     return {}

# @hookspec
# def get_ADMINDATAVIEWS():
#     return {}

# @hookspec
# def get_QUEUES():
#     return {}


##############################################################
# provided by abx.pydantic_pkgr.hookspec:
# @hookspec
# def get_BINARIES() -> Dict[str, BaseBinary]:
#     return {}

# @hookspec
# def get_BINPROVIDERS() -> Dict[str, BaseBinProvider]:
#     return {}



__package__ = 'abx.archivebox'

import abx


class BaseReplayer:
    """Describes how to render an ArchiveResult in several contexts"""
    
    url_pattern: str = '*'

    row_template: str = 'plugins/generic_replayer/templates/row.html'
    embed_template: str = 'plugins/generic_replayer/templates/embed.html'
    fullpage_template: str = 'plugins/generic_replayer/templates/fullpage.html'

    # row_view: LazyImportStr = 'plugins.generic_replayer.views.row_view'
    # embed_view: LazyImportStr = 'plugins.generic_replayer.views.embed_view'
    # fullpage_view: LazyImportStr = 'plugins.generic_replayer.views.fullpage_view'
    # icon_view: LazyImportStr = 'plugins.generic_replayer.views.get_icon'
    # thumbnail_view: LazyImportStr = 'plugins.generic_replayer.views.get_icon'

    @abx.hookimpl
    def get_REPLAYERS(self):
        return [self]

    # TODO: add hookimpl methods for get_row_template, get_embed_template, get_fullpage_template, etc...



__package__ = "abx.archivebox"

import os
from typing import Optional, cast
from typing_extensions import Self

from pydantic import validate_call
from pydantic_pkgr import (
    Binary,
    BinProvider,
    BinProviderName,
    AptProvider,
    BrewProvider,
    EnvProvider,
)

from archivebox.config.permissions import ARCHIVEBOX_USER

import abx


class BaseBinProvider(BinProvider):
    
    # TODO: add install/load/load_or_install methods as abx.hookimpl methods
    
    @property
    def admin_url(self) -> str:
        # e.g. /admin/environment/binproviders/NpmBinProvider/   TODO
        return "/admin/environment/binaries/"

    @abx.hookimpl
    def get_BINPROVIDERS(self):
        return [self]

class BaseBinary(Binary):
    # TODO: formalize state diagram, final states, transitions, side effects, etc.

    @staticmethod
    def symlink_to_lib(binary, bin_dir=None) -> None:
        from archivebox.config.common import STORAGE_CONFIG
        bin_dir = bin_dir or STORAGE_CONFIG.LIB_DIR / 'bin'
        
        if not (binary.abspath and os.access(binary.abspath, os.R_OK)):
            return
        
        try:
            bin_dir.mkdir(parents=True, exist_ok=True)
            symlink = bin_dir / binary.name
            symlink.unlink(missing_ok=True)
            symlink.symlink_to(binary.abspath)
            symlink.chmod(0o777)   # make sure its executable by everyone
        except Exception as err:
            # print(f'[red]:warning: Failed to symlink {symlink} -> {binary.abspath}[/red] {err}')
            # not actually needed, we can just run without it
            pass
        
    @validate_call
    def load(self, fresh=False, **kwargs) -> Self:
        from archivebox.config.common import STORAGE_CONFIG
        if fresh:
            binary = super().load(**kwargs)
            self.symlink_to_lib(binary=binary, bin_dir=STORAGE_CONFIG.LIB_DIR / 'bin')
        else:
            # get cached binary from db
            try:
                from machine.models import InstalledBinary
                installed_binary = InstalledBinary.objects.get_from_db_or_cache(self)    # type: ignore
                binary = InstalledBinary.load_from_db(installed_binary)
            except Exception:
                # maybe we are not in a DATA dir so there is no db, fallback to reading from fs
                # (e.g. when archivebox version is run outside of a DATA dir)
                binary = super().load(**kwargs)
        return cast(Self, binary)
    
    @validate_call
    def install(self, **kwargs) -> Self:
        from archivebox.config.common import STORAGE_CONFIG
        binary = super().install(**kwargs)
        self.symlink_to_lib(binary=binary, bin_dir=STORAGE_CONFIG.LIB_DIR / 'bin')
        return binary
    
    @validate_call
    def load_or_install(self, fresh=False, **kwargs) -> Self:
        from archivebox.config.common import STORAGE_CONFIG
        try:
            binary = self.load(fresh=fresh)
            if binary and binary.version:
                self.symlink_to_lib(binary=binary, bin_dir=STORAGE_CONFIG.LIB_DIR / 'bin')
                return binary
        except Exception:
            pass
        return self.install(**kwargs)
    
    @property
    def admin_url(self) -> str:
        # e.g. /admin/environment/config/LdapConfig/
        return f"/admin/environment/binaries/{self.name}/"

    @abx.hookimpl
    def get_BINARIES(self):
        return [self]


class AptBinProvider(AptProvider, BaseBinProvider):
    name: BinProviderName = "apt"
    
class BrewBinProvider(BrewProvider, BaseBinProvider):
    name: BinProviderName = "brew"
    
class EnvBinProvider(EnvProvider, BaseBinProvider):
    name: BinProviderName = "env"
    
    euid: Optional[int] = ARCHIVEBOX_USER

apt = AptBinProvider()
brew = BrewBinProvider()
env = EnvBinProvider()
